@@ -8,21 +8,6 @@ use Fcntl;   # For O_RDWR, O_CREAT, etc.
 use SDBM_File;
 use FreezeThaw;
 
-sub _dbfile() {
-    # Base SDBM filename on last component of configuration directory
-    my $tag = $ENV{MAsessiontag};
-    if (!defined $tag) {
-	my $dir = $ENV{MAconfigdir} or
-	    die "no MAsessiontag or MAconfigdir defined in environment";
-	my @comp = split('/', $dir);
-	for (my $i = @comp-1; $i >= 0; $i--) {
-	    $tag = $comp[$i];
-	    last if $tag ne "admin" && $tag ne "mkadmin";
-	}
-    }
-    return "/tmp/mkadmin-sessions-$tag.sdbm";
-}
-
 sub new {
     my $class = shift();
 
@@ -45,15 +30,11 @@ sub find {
     my $this = shift();
     my($cookie) = @_;
 
-    my %h;
-    my $dbfile = _dbfile();
-    if (!tie(%h, "SDBM_File", $dbfile, O_RDONLY, 0666)) {
-	return undef if $! =~ /No such file/i;
-	die "Couldn't read SDBM file '$dbfile': $!";
-    }
-
-    my $val = $h{$cookie};
-    untie %h;
+    my $val;
+    _with_dbfile(0, sub {
+	my($href) = @_;
+	$val = $href->{$cookie};
+    });
 
     return undef if !defined $val;
     my($session) = FreezeThaw::thaw($val);
@@ -74,18 +55,77 @@ sub update {
     }
 
     if ($count > 0) {
-	my %h;
-	my $dbfile = _dbfile();
-	# For some reason, O_WRONLY|O_CREAT fails with ENOENT
-	tie(%h, "SDBM_File", $dbfile, O_RDWR|O_CREAT, 0666)
-	    or die "Couldn't write SDBM file '$dbfile': $!";
-
-	my $cookie = $this->cookie();
-	$h{$cookie} = FreezeThaw::freeze($this);
-	untie %h;
+	_with_dbfile(1, sub {
+	    my($href) = @_;
+	    my $cookie = $this->cookie();
+	    $href->{$cookie} = FreezeThaw::freeze($this);
+	});
     }
 
     return $count
 }
 
+sub delete {
+    my $this = shift();
+
+    _with_dbfile(1, sub {
+	my($href) = @_;
+	my $cookie = $this->cookie();
+	delete $href->{$cookie};
+    });
+}
+
+# Accepts a list of candidate links, each of them an array-reference
+# whose 0th element is a session identifier. Returns a new list
+# consisting of those elements for which the session identifier
+# corresponds to a session that actually exists.
+#
+sub filter_links {
+    my(@list) = @_;
+    my @res;
+
+    _with_dbfile(0, sub {
+	my($href) = @_;
+	foreach my $ref (@list) {
+	    my $sessid = $ref->[0];
+	    push @res, $ref if exists $href->{$sessid};
+	}
+    });
+
+    return @res;
+}
+
+sub _with_dbfile {
+    my($readWrite, $callback) = @_;
+    # For some reason, O_WRONLY|O_CREAT fails with ENOENT
+    my $flags = $readWrite ? (O_RDWR|O_CREAT) : O_RDONLY;
+
+    my $tag = $ENV{MAsessiontag};
+    if (!defined $tag) {
+	my $dir = $ENV{MAconfigdir} or
+	    die "no MAsessiontag or MAconfigdir defined in environment";
+	my @comp = split('/', $dir);
+	for (my $i = @comp-1; $i >= 0; $i--) {
+	    $tag = $comp[$i];
+	    last if $tag ne "admin" && $tag ne "mkadmin";
+	}
+    }
+    my $dbfile = "/tmp/mkadmin-sessions-$tag.sdbm";
+
+    my %h;
+    eval {
+	tie(%h, "SDBM_File", $dbfile, $flags, 0666)
+	    or die "Couldn't write SDBM file '$dbfile': $!";
+	&$callback(\%h);
+    }; if ($@ && !$readWrite) {
+	# Doesn't matter that we're not really tied, since we're not making updates
+	&$callback({});
+    } elsif ($@) {
+	die $@;
+    }
+
+    untie %h;
+    
+}
+
 1;
